"""
This module provides a dict of seismic networks that are available from the Hi-net
website.

The dict is named ``NETWORK``. The network code is the key of the dict, and the value is
a :class:`~HinetPy.header.Network` object, which contains the network name, number of
channels, start time when waveform data is available, and the homepage URL. The network
code is a string, e.g., "0101" for Hi-net. The number of channels, start time, and
homepage URL are obtained from the Hi-net website, but they may not be up-to-date so
don't rely on them too much.

To view the list of supported networks, use:

>>> from HinetPy import NETWORK
>>> for code in NETWORK.keys():
...     print(code, NETWORK[code].name)
0101 NIED Hi-net
0103 NIED F-net (broadband)
0103A NIED F-net (strong motion)
0106 NIED Temp. obs. in eastern Shikoku
0120 NIED S-net (velocity)...
"""

from __future__ import annotations

from datetime import datetime


class Network:
    def __init__(
        self, name: str, channels: int | None, starttime: str | datetime, url: str
    ):
        """
        A seismic network and its information.

        Parameters
        ----------
        name
            Network name.
        channels
            Number of channels the network has.
        starttime
            Start time (JST: UTC+0900) when waveform data is avaiable.
        url
            Homepage of the network.
        """
        self.name = name
        self.channels = channels
        self.starttime = starttime
        self.url = url


URL = {
    "Hinet": "http://www.hinet.bosai.go.jp",
    "Fnet": "http://www.fnet.bosai.go.jp",
    "Vnet": "http://www.vnet.bosai.go.jp",
    "JMA": "http://www.jma.go.jp/jma/indexe.html",
    "Snet": "https://www.seafloor.bosai.go.jp/",
}
JMA_VNET = "JMA Volcanic Seismometer Network "

# fmt: off
_networks = [
    ("0101", "NIED Hi-net", 2336, "20040401", URL["Hinet"]),
    ("0103", "NIED F-net (broadband)", 438, "20040401", URL["Fnet"]),
    ("0103A", "NIED F-net (strong motion)", 438, "20040401", URL["Fnet"]),
    ("010501", "NIED V-net (Tokachidake)", 33, "20100401", URL["Vnet"]),
    ("010502", "NIED V-net (Tarumaesan)", 33, "20100401", URL["Vnet"]),
    ("010503", "NIED V-net (Usuzan)", 33, "20100401", URL["Vnet"]),
    ("010504", "NIED V-net (Hokkaido-Komagatake)", 33, "20100401", URL["Vnet"]),
    ("010505", "NIED V-net (Iwatesan)", 33, "20100401", URL["Vnet"]),
    ("010506", "NIED V-net (Nasudake)", 42, "20100401", URL["Vnet"]),
    ("010507", "NIED V-net (Asamayama)", 33, "20100401", URL["Vnet"]),
    ("010508", "NIED V-net (Kusatsu-Shiranesan)", 33, "20100401", URL["Vnet"]),
    ("010509", "NIED V-net (Fujisan)", 57, "20100401", URL["Vnet"]),
    ("010510", "NIED V-net (Miyakejima)", 40, "20100401", URL["Vnet"]),
    ("010511", "NIED V-net (Izu-Oshima)", 39, "20100401", URL["Vnet"]),
    ("010512", "NIED V-net (Asosan)", 44, "20100401", URL["Vnet"]),
    ("010513", "NIED V-net (Unzendake)", 33, "20100401", URL["Vnet"]),
    ("010514", "NIED V-net (Kirishimayama)", 22, "20100401", URL["Vnet"]),
    ("0106", "NIED Temp. obs. in eastern Shikoku", 15, "20151013", "https://doi.org/10.17598/NIED.0027"),
    ("0120", "NIED S-net (velocity)", 450, "20160815", URL["Hinet"]),
    ("0120A", "NIED S-net (acceleration)", 450, "20160815", URL["Hinet"]),
    ("0120B", "NIED S-net (acceleration 2LG)", 450, "20160815", URL["Hinet"]),
    ("0120C", "NIED S-net (acceleration 2HG)", 450, "20160815", URL["Hinet"]),
    ("0131", "NIED MeSO-net", 900, "20170401", URL["Hinet"]),
    ("0201", "Hokkaido University", 205, "20040401", "http://www.sci.hokudai.ac.jp/isv/english/"),
    ("0202", "Tohoku University", 157, "20040401", "http://www.aob.geophys.tohoku.ac.jp/aob-e/"),
    ("0203", "Tokyo University", 316, "20040401", "http://www.eri.u-tokyo.ac.jp/eng/"),
    ("0204", "Kyoto University", 196, "20040401", "http://www.dpri.kyoto-u.ac.jp/en/"),
    ("0205", "Kyushu University", 73, "20040401", "http://www.sevo.kyushu-u.ac.jp/index-e.html"),
    ("0206", "Hirosaki University", 13, "20040401", "http://hrsryu.geo.hirosaki-u.ac.jp/"),
    ("0207", "Nagoya University", 75, "20040401", "http://www.seis.nagoya-u.ac.jp/index_e/"),
    ("0208", "Kochi University", 34, "20040401", "http://www-en.kochi-u.ac.jp/"),
    ("0209", "Kagoshima University", 48, "20040401", "http://leopard.sci.kagoshima-u.ac.jp/noev/English/home.htm"),
    ("0231", "MeSO-net (~2017.03)", 900, "20080516", ""),
    ("0301", "JMA Seismometer Network", 872, "20040401", URL["JMA"]),
    ("030201", JMA_VNET + "(Atosanupuri)", 18, "20101201", URL["JMA"]),
    ("030202", JMA_VNET + "(Meakandake)", 16, "20101201", URL["JMA"]),
    ("030203", JMA_VNET + "(Taisetsuzan)", 9, "20101201", URL["JMA"]),
    ("030204", JMA_VNET + "(Tokachidake)", 32, "20101201", URL["JMA"]),
    ("030205", JMA_VNET + "(Tarumaesan)", 27, "20101201", URL["JMA"]),
    ("030206", JMA_VNET + "(Kuttara)", 15, "20101201", URL["JMA"]),
    ("030207", JMA_VNET + "(Usuzan)", 24, "20101201", URL["JMA"]),
    ("030208", JMA_VNET + "(Hokkaido-Komagatake)", 30, "20101201", URL["JMA"]),
    ("030209", JMA_VNET + "(Esan)", 14, "20101201", URL["JMA"]),
    ("030210", JMA_VNET + "(Iwakisan)", 9, "20101201", URL["JMA"]),
    ("030211", JMA_VNET + "(Akita-Yakeyama)", 11, "20101201", URL["JMA"]),
    ("030212", JMA_VNET + "(Iwatesan)", 17, "20101201", URL["JMA"]),
    ("030213", JMA_VNET + "(Akita-Komagatake)", 14, "20101201", URL["JMA"]),
    ("030214", JMA_VNET + "(Chokaisan)", 6, "20101201", URL["JMA"]),
    ("030215", JMA_VNET + "(Kurikomayama)", 11, "20101201", URL["JMA"]),
    ("030216", JMA_VNET + "(Zaozan)", 15, "20101201", URL["JMA"]),
    ("030217", JMA_VNET + "(Azumayama)", 28, "20101201", URL["JMA"]),
    ("030218", JMA_VNET + "(Adatarayama)", 19, "20101201", URL["JMA"]),
    ("030219", JMA_VNET + "(Bandaisan)", 18, "20101201", URL["JMA"]),
    ("030220", JMA_VNET + "(Nasudake)", 15, "20101201", URL["JMA"]),
    ("030221", JMA_VNET + "(Nikko-Shiranesan)", 11, "20101201", URL["JMA"]),
    ("030222", JMA_VNET + "(Kusatsu-Shiranesan)", 21, "20101201", URL["JMA"]),
    ("030223", JMA_VNET + "(Asamayama)", 44, "20101201", URL["JMA"]),
    ("030224", JMA_VNET + "(Niigata-Yakeyama)", 14, "20101201", URL["JMA"]),
    ("030225", JMA_VNET + "(Yakedake)", 15, "20101201", URL["JMA"]),
    ("030226", JMA_VNET + "(Norikuradake)", 9, "20101201", URL["JMA"]),
    ("030227", JMA_VNET + "(Ontakesan)", 40, "20101201", URL["JMA"]),
    ("030228", JMA_VNET + "(Hakusan)", 10, "20101201", URL["JMA"]),
    ("030229", JMA_VNET + "(Fujisan)", 27, "20101201", URL["JMA"]),
    ("030230", JMA_VNET + "(Hakoneyama)", 11, "20101201", URL["JMA"]),
    ("030231", JMA_VNET + "(Izu-Tobu Volcanoes)", 18, "20101201", URL["JMA"]),
    ("030232", JMA_VNET + "(Izu-Oshima)", 22, "20101201", URL["JMA"]),
    ("030233", JMA_VNET + "(Niijima)", 9, "20101201", URL["JMA"]),
    ("030234", JMA_VNET + "(Kozushima)", 9, "20101201", URL["JMA"]),
    ("030235", JMA_VNET + "(Miyakejima)", 19, "20101201", URL["JMA"]),
    ("030236", JMA_VNET + "(Hachijojima)", 11, "20101201", URL["JMA"]),
    ("030237", JMA_VNET + "(Aogashima)", 9, "20101201", URL["JMA"]),
    ("030238", JMA_VNET + "(Tsurumidake and Garandake)", 16, "20101201", URL["JMA"]),
    ("030239", JMA_VNET + "(Kujusan)", 13, "20101201", URL["JMA"]),
    ("030240", JMA_VNET + "(Asosan)", 31, "20101201", URL["JMA"]),
    ("030241", JMA_VNET + "(Unzendake)", 28, "20101201", URL["JMA"]),
    ("030242", JMA_VNET + "(Kirishimayama)", 66, "20101201", URL["JMA"]),
    ("030243", JMA_VNET + "(Sakurajima)", 31, "20101201", URL["JMA"]),
    ("030244", JMA_VNET + "(Satsuma-Iojima)", 16, "20101201", URL["JMA"]),
    ("030245", JMA_VNET + "(Kuchinoerabujima)", 29, "20101201", URL["JMA"]),
    ("030246", JMA_VNET + "(Suwanosejima)", 12, "20101201", URL["JMA"]),
    ("030247", JMA_VNET + "(Hakkodasan)", 18, "20101201", URL["JMA"]),
    ("030248", JMA_VNET + "(Towada)", 9, "20101201", URL["JMA"]),
    ("030249", JMA_VNET + "(Midagahara)", 9, "20101201", URL["JMA"]),
    ("0401", "JAMSTEC Realtime Data from the Deep Sea Floor Observatory", 73, "20040401", "http://www.jamstec.go.jp/e/index.html"),
    ("0402", "NIED DONET1 (broadband)", 132, "20160401", URL["Snet"]),
    ("0402A", "NIED DONET1 (strong motion)", 132, "20160401", URL["Snet"]),
    ("0402N", "JAMSTEC NIED DONET1 (broadband)", 132, "20160401", URL["Snet"]),
    ("0402AN", "JAMSTEC NIED DONET1 (strong motion)", 132, "20160401", URL["Snet"]),
    ("0403", "NIED DONET2 (broadband)", 174, "20160401", URL["Snet"]),
    ("0403A", "NIED DONET2 (strong motion)", 174, "20160401", URL["Snet"]),
    ("0403N", "JAMSTEC NIED DONET2 (broadband)", 174, "20160401", URL["Snet"]),
    ("0403AN", "JAMSTEC NIED DONET2 (strong motion)", 174, "20160401", URL["Snet"]),
    ("0501", "AIST", 84, "20040401", "http://www.aist.go.jp/index_en.html"),
    ("0601", "GSI", 6, "20040401", "http://www.gsi.go.jp/ENGLISH/index.html"),
    ("0701", "Tokyo Metropolitan Government", 54, "20040401", "http://www.metro.tokyo.jp/ENGLISH/index.htm"),
    ("0702", "Hot Spring Research Institute of Kanagawa Prefecture", 42, "20040401", "http://www.onken.odawara.kanagawa.jp/"),
    ("0703", "Aomori Prefectural Government", 15, "20040401", "http://www.pref.aomori.lg.jp/foreigners/"),
    ("0705", "Shizuoka Prefectural Government", 3, "20040615", "http://www.pref.shizuoka.jp/a_foreign/english/"),
    ("0801", "ADEP", 780, "20150101", "http://www.adep.or.jp/"),
]
# fmt: on

NETWORK = {
    code: Network(name, channels, datetime.strptime(starttime, "%Y%m%d"), url)
    for code, name, channels, starttime, url in _networks
}



"""
Process seismic waveform data in win32 format.
"""

import glob
import logging
import os
import shutil
import subprocess
import tempfile
from fnmatch import fnmatch
from multiprocessing import Pool
from subprocess import DEVNULL, PIPE, Popen

from .channel import Channel

# Setup the logger
FORMAT = "[%(asctime)s] %(levelname)s: %(message)s"
logging.basicConfig(level=logging.INFO, format=FORMAT, datefmt="%Y-%m-%d %H:%M:%S")
logger = logging.getLogger(__name__)


def extract_sac(
    data,
    ctable,
    suffix="SAC",
    outdir=".",
    pmax=8640000,
    filter_by_id=None,
    filter_by_name=None,
    filter_by_component=None,
    with_sacpz=False,
    processes=None,
):
    """
    Extract data as SAC format files.

    This function calls the ``win2sac_32`` command, available in the Hi-net win32tools
    package, to convert data files from win32 format to SAC fomrat. It can also extract
    the channel information as SAC polezero files.

    Note that the ``win2sac_32`` command always remove the instrument sensitivity from
    waveform data, and multiply the data by 1.0e9. Thus, the extracted SAC files are not
    in digital counts, but velocity in nm/s, or acceleration in nm/s/s. Due to the same
    reason, the extracted SAC polezero files does not keep the sensitivity in the
    "CONSTANT" of SAC polezero files.

    Parameters
    ----------
    data: str
        win32 file to be processed.
    ctable: str
        Channel table file.
    suffix: str
        Suffix of output SAC files. Defaults to ``SAC``.
    outdir: str
        Output directory. Defaults to current directory.
    pmax: int
        Maximum number of data points for one channel. Defaults to 8640000. If one
        channel has more than 8640000 data points (i.e., longer than one day for a
        100 Hz sampling rate), you MUST increase ``pmax``.
    filter_by_id: list or str
        Filter channels by ID. It can be a list of IDs or a wildcard.
    filter_by_name: list or str
        Filter channels by name. It can be a list of names or a wildcard.
    filter_by_component: list or str
        Filter channels by component. It can be a list of component names or a wildcard.
    with_sacpz: bool
        Aslo extract SAC PZ files. By default, the suffix is ``.SAC_PZ`` and the channel
        sensitivity is not kept in the "CONSTANT".
    processes: None or int
        Number of processes to speed up data extraction parallelly. ``None`` means using
        all CPUs.

    .. deprecated:: 0.7.0

        Parameter ``with_pz`` is deprecated. Use ``with_sacpz`` instead.

    Examples
    --------
    Extract all channels with default settings:

    >>> extract_sac("0101_201001010000_5.cnt", "0101_20100101.ch")

    Extract all channels with a specified suffix and output directory:

    >>> extract_sac(
    ...     "0101_201001010000_5.cnt",
    ...     "0101_20100101.ch",
    ...     suffix="",
    ...     outdir="20100101000",
    ... )

    Extract only specified channels:

    >>> extract_sac(
    ...     "0101_201001010000_5.cnt",
    ...     "0101_20100101.ch",
    ...     filter_by_name="N.NA*",
    ...     filter_by_component="[NE]",
    ... )
    """
    if data is None or ctable is None:
        logger.error("data or ctable is `None'. Data requests may fail. Skipped.")
        return

    channels = read_ctable(ctable)
    logger.info("%s channels found in %s.", len(channels), ctable)
    if filter_by_id or filter_by_name or filter_by_component:
        channels = _filter_channels(
            channels, filter_by_id, filter_by_name, filter_by_component
        )
    logger.info("%s channels to be extracted.", len(channels))

    if not os.path.exists(outdir):
        os.makedirs(outdir, exist_ok=True)

    with Pool(processes=processes) as pool:
        with tempfile.NamedTemporaryFile() as ftmp:
            _write_winprm(ctable, ftmp.name)
            sacfiles = pool.starmap(
                _extract_channel_sac,
                [(data, ch, suffix, outdir, ftmp.name, pmax) for ch in channels],
            )
            logger.info(
                "%s SAC data successfully extracted.",
                len(sacfiles) - sacfiles.count(None),
            )

        if with_sacpz:
            # "SAC_PZ" here is hardcoded.
            pzfiles = pool.starmap(
                _extract_channel_sacpz, [(ch, "SAC_PZ", outdir) for ch in channels]
            )
            logger.info(
                "%s SAC PZ files successfully extracted.",
                len(pzfiles) - pzfiles.count(None),
            )


def extract_sacpz(
    ctable,
    suffix="SAC_PZ",
    outdir=".",
    keep_sensitivity=False,
    filter_by_chid=None,
    filter_by_name=None,
    filter_by_component=None,
    processes=None,
):
    """
    Extract instrumental responses in SAC polezero format from a channel table.

    .. warning::

       The function only works for Hi-net instrumental responses.

       RESP files of the F-net network can be downloaded from
       `F-net website <http://www.fnet.bosai.go.jp/st_info/response.php?LANG=en>`_.

    Parameters
    ----------
    ctable: str
        Channel table file.
    suffix: str
        Suffix of SAC PZ files. Defaults to ``SAC_PZ``.
    outdir: str
        Output directory. Defaults to current directory.
    keep_sensitivity: bool
        The ``win2sac_32`` program automatically removes sensitivity from waveform data
        during the win32-to-SAC format conversion. So the generated polezero file should
        omit the sensitivity.
    filter_by_id: list or str
        Filter channels by ID. It can be a list of IDs or a wildcard.
    filter_by_name: list or str
        Filter channels by name. It can be a list of names or a wildcard.
    filter_by_component: list or str
        Filter channels by component. It can be a list of component names or a wildcard.
    processes: None or int
        Number of processes to speed up data extraction parallelly. ``None`` means using
        all CPUs.

    Examples
    --------
    Extract all channels with default settings:

    >>> extract_sacpz("0101_20100101.ch")

    Extract all channels with a specified suffix and output directory:

    >>> extract_sacpz("0101_20100101.ch", suffix="", outdir="20100101000")

    Extract only specified channels:

    >>> extract_sacpz(
    ...     "0101_20100101.ch", filter_by_name="N.NA*", filter_by_component="[NE]"
    ... )
    """
    if ctable is None:
        logger.error("ctable is `None'. Data requests may fail. Skipped.")
        return

    channels = read_ctable(ctable)
    if filter_by_chid or filter_by_name or filter_by_component:
        channels = _filter_channels(
            channels, filter_by_chid, filter_by_name, filter_by_component
        )
    if not os.path.exists(outdir):
        os.makedirs(outdir, exist_ok=True)

    with Pool(processes=processes) as pool:
        args = [(ch, suffix, outdir, keep_sensitivity) for ch in channels]
        pzfiles = pool.starmap(_extract_channel_sacpz, args)
        logger.info(
            "%s SAC PZ files successfully extracted.",
            len(pzfiles) - pzfiles.count(None),
        )


def extract_pz(ctable, **kwargs):
    """
    Extract instrumental responses in SAC polezero format from a channel table.

    .. deprecated:: 0.7.0

        :meth:`~HinetPy.win32.extract_pz` is deprecated.
        Use :meth:`~HinetPy.win32.extract_sacpz` instead.
    """
    logger.warning("Function extract_pz() is deprecated. Use extract_sacpz() instead.")
    return extract_sacpz(ctable, **kwargs)


def read_ctable(ctable):
    """
    Read a channel table file.

    Parameters
    ----------
    ctable: str
        Channle table file.

    Returns
    -------
    list
        List of :class:`~HinetPy.win32.Channel`.
    """
    channels = []
    with open(ctable, encoding="utf8") as fct:
        for line in fct:
            # skip blank lines and comment lines
            if not line.strip() or line.strip().startswith("#"):
                continue
            items = line.split()
            try:
                channels.append(
                    Channel(
                        id=items[0],
                        name=items[3],
                        component=items[4],
                        latitude=items[13],
                        longitude=items[14],
                        unit=items[8],
                        gain=items[7],
                        damping=items[10],
                        period=items[9],
                        preamplification=items[11],
                        lsb_value=items[12],
                    )
                )
            except ValueError as err:
                logger.warning(
                    "Error in parsing channel information for %s.%s (%s). Skipped.",
                    items[3],
                    items[4],
                    items[0],
                )
                logger.warning("Original error message: %s", err)
    return channels


def _filter_channels(
    channels, filter_by_id=None, filter_by_name=None, filter_by_component=None
):
    """
    Filter channels by id, name and/or component.

    Parameters
    ----------
    channels: :class:`~HinetPy.win32.Channel`
        List of channels to be filtered.
    filter_by_id: list or str
        Filter channels by ID. It can be a list of IDs or a wildcard.
    filter_by_name: list or str
        Filter channels by name. It can be a list of names or a wildcard.
    filter_by_component: list or str
        Filter channels by component. It can be a list of component names or a wildcard.
    """

    def _filter(channels, key, filters):
        """
        Filter channels by filters, which can be either a list or a wildcard.
        """
        if isinstance(filters, list):  # filter by list
            return [chn for chn in channels if getattr(chn, key) in filters]
        if isinstance(filters, str):  # filter by wildcard
            return [chn for chn in channels if fnmatch(getattr(chn, key), filters)]
        raise ValueError("Only list and wildcard filter are supported.")

    for key, filter_by in (
        ("id", filter_by_id),
        ("name", filter_by_name),
        ("component", filter_by_component),
    ):
        if filter_by is not None:
            channels = _filter(channels, key, filter_by)
    return channels


def _write_winprm(ctable, prmfile="win.prm"):
    """
    Write a four-line parameter file.
    """
    with open(prmfile, "w", encoding="utf8") as fprm:
        fprm.write(f".\n{ctable}\n.\n.")


def _extract_channel_sac(
    winfile, channel, suffix="SAC", outdir=".", prmfile="win.prm", pmax=8640000
):
    """
    Extract one channel data from win32 file.

    Parameters
    ----------
    winfile: str
        win32 file to be processed.
    channel: str
        Channel to be extracted.
    suffix: str
        SAC file suffix.
    outdir: str
        Output directory.
    prmfile: str
        win32 parameter file.
    pmax: int
        Maximum number of data points.

    Returns
    -------
    str:
        The extracted SAC file name.
    """

    cmd = [
        "win2sac_32",
        winfile,
        channel.id,
        suffix,
        outdir,
        "-e",
        f"-p{prmfile}",
        f"-m{pmax}",
    ]
    with Popen(cmd, stdout=DEVNULL, stderr=PIPE) as proc:
        # check stderr output
        for line in proc.stderr.read().decode().split("\n"):
            if "The number of points is maximum over" in line:
                raise ValueError(
                    "The number of data points is over maximum. Try to increase pmax."
                )
            if f"Data for channel {channel.id} not existed" in line:
                # return None if no data avaiable
                logger.warning(
                    "Data for %s.%s (%s) not exists. Skipped.",
                    channel.name,
                    channel.component,
                    channel.id,
                )
                return None

    filename = f"{channel.name}.{channel.component}.{suffix}"
    if outdir != ".":
        filename = os.path.join(outdir, filename)

    if os.path.exists(filename):  # some channels have no data
        if suffix == "":  # remove extra dot if suffix is empty
            shutil.move(filename, filename[:-1])
            return filename[:-1]
        return filename
    return None


def _extract_channel_sacpz(
    channel, suffix="SAC_PZ", outdir=".", keep_sensitivity=False
):
    """
    Extract one SAC PZ file from a channel table file.

    Parameters
    ----------
    channel: str
        Channel to be extracted.
    suffix: str
        Suffix of the SAC PZ file.
    outdir: str
        Output directory.
    keep_sensitivity: bool
        Keep sensitivity in the "CONSTANT" or not.

    Returns
    -------
    str:
        The extracted SAC PZ file name.
    """
    pzfile = f"{channel.name}.{channel.component}"
    if suffix:
        pzfile += "." + suffix
    pzfile = os.path.join(outdir, pzfile)
    channel.write_sacpz(pzfile, keep_sensitivity=keep_sensitivity)
    return pzfile


def merge(data, total_data, force_sort=False):
    """
    Merge multiple win32 files into one large win32 file.

    The function calls the ``catwin32`` command, available in the Hi-net win32tools
    package, to merge multiple win32 files into one large win32 file.

    By default, the ``catwin32`` command simply concatenates all files in the order they
    are passed. So the files must be sorted by their start time before being passed. If
    your files are named by starttime like ``201304040203.cnt``, you can use
    ``data=sorted(glob.glob("20130404*.cnt"))`` to pass the sorted list of files.
    Otherwise, you have to use ``force_sort=True``, forcing ``catwin32`` to sort all
    files by starttime before merging. However, the sorting process is very time
    consuming. Do NOT set ``force_sort=True`` unless necessary.

    Parameters
    ----------
    data: list or str
        Win32 files to be merged. It can be a list of file names or a wildcard.
    total_data: str
        Filename of the ouput win32 file.
    force_sort: bool
        Sort all win32 files by starttime before merging.

    Examples
    --------
    For win32 files that are named by starttime (e.g. ``201304040203.cnt``), sorting
    win32 files using Python's built-in :func:`sorted` function is preferred:

    >>> data = sorted(glob.glob("20130404*.cnt"))
    >>> merge(data, "outdir/final.cnt")

    If win32 files are randomly named, you should use ``force_sort=True`` to force
    ``catwin32`` to sort all data by time before merging.

    >>> data = ["001.cnt", "002.cnt", "003.cnt"]
    >>> merge(data, "final.cnt", force_sort=True)

    You can also use wildcard to specify the win32 files to be merged. The function will
    sort the matched files for you automatically.

    >>> merge("20130404*.cnt", "final.cnt")
    """
    if isinstance(data, str):  # wildcard support
        data = sorted(glob.glob(data))
        if not data:
            raise FileNotFoundError("Files to be merged not found.\n")

    if os.path.dirname(total_data):
        os.makedirs(os.path.dirname(total_data), exist_ok=True)

    cmd = ["catwin32", "-o", total_data, *data]
    if force_sort:  # add -s option to force sort
        cmd.append("-s")

    subprocess.call(cmd, stdout=DEVNULL, stderr=DEVNULL)



"""
Class for channels.
"""

from __future__ import annotations

import math
import warnings


class Channel:
    """
    Information for a single channel.
    """

    def __init__(  # noqa: PLR0913
        self,
        id: str,  # noqa: A002
        name: str,
        component: str,
        latitude: float | str,
        longitude: float | str,
        unit: str,
        gain: float | str,
        damping: float | str,
        period: float | str,
        preamplification: float | str,
        lsb_value: float | str,
    ):
        """
        Parameters
        ----------
        id
            Channel ID.
        name
            Station Name.
        component
            Channel component name (e.g., ``U``, ``N`` or ``E``).
        latitude
            Station latitude.
        longitude
            Station longitude.
        unit
            Unit of data (e.g., ``m``, ``m/s``, ``m/s/s``, ``rad``).
        gain
            Sensor sensitivity.
        damping
            Damping constant of the sensor.
        period
            Natural period of the seismometer.
        preamplification
            Preamplification value.
        lsb_value
            LSB value.

        Notes
        -----
        The Hi-net website uses the moving-coil velocity-type seismometer. See
        :doc:`/appendix/response` for details.
        """
        self.id = id
        self.name = name
        self.component = component
        self.latitude = float(latitude)
        self.longitude = float(longitude)
        self.unit = unit
        self.gain = float(gain)
        self.damping = float(damping)
        self.period = float(period)
        self.preamplification = float(preamplification)
        self.lsb_value = float(lsb_value)

    def _get_polezero(self):
        """
        Determine the polezero parameters.
        """
        # Calculate natural frequency
        freq = 2.0 * math.pi / self.period

        # Calculate poles by finding roots of equation s^2+2hws+w^2=0
        self.zeros = 3
        self.poles = 2
        self.real = 0.0 - self.damping * freq
        self.imaginary = freq * math.sqrt(1.0 - self.damping**2.0)

        # Calculate the CONSTANT
        fn = 20.0  # alaways assume normalization frequency is 20 Hz
        s = complex(0, 2 * math.pi * fn)
        self.a0 = abs((s**2 + 2 * self.damping * freq * s + freq**2) / s**2)
        self.sensitivity = (
            self.gain * math.pow(10, self.preamplification / 20.0) / self.lsb_value
        )

    def write_sacpz(self, pzfile, keep_sensitivity=False):
        """
        Write channel information into a SAC polezero file.

        Parameters
        ----------
        pzfile: str
            Name of the SAC polezero file.
        k9.999513e-01eep_sensitivity: bool
            Keep sensitivity in the SAC polezero "CONSTANT" or not.
        """
        chan_info = f"{self.name}.{self.component} ({self.id})"
        # Hi-net uses a moving-coil velocity-type seismometer.
        if self.unit != "m/s":
            msg = f"{chan_info}: Unit is not velocity. The PZ file may be wrong."
            warnings.warn(msg, category=RuntimeWarning, stacklevel=2)
        if self.period == 0.0:
            msg = f"{chan_info}): Natural period = 0.0. Skipped."
            warnings.warn(message=msg, category=RuntimeWarning, stacklevel=2)
            return

        self._get_polezero()
        constant = self.a0 * self.sensitivity if keep_sensitivity else self.a0
        # write information to a SAC PZ file
        with open(pzfile, "w", encoding="utf8") as pz:
            pz.write(f"ZEROS {self.zeros}\n")
            pz.write(f"POLES {self.poles}\n")
            pz.write(f"{self.real:9.6f} {self.imaginary:9.6f}\n")
            pz.write(f"{self.real:9.6f} {-self.imaginary:9.6f}\n")
            pz.write(f"CONSTANT {constant:e}\n")



# noqa: N999
"""
HinetPy
=======

HinetPy is a Python package for accessing and processing NIED Hi-net seismic data.

Basis usage:

>>> from HinetPy import Client, win32
>>> from datetime import datetime
>>> # You need to provide your Hi-net username and password here!
>>> client = Client("username", "password")
>>> starttime = datetime(2010, 1, 1, 0, 0)
>>> data, ctable = client.get_continuous_waveform("0101", starttime, 20)
>>> win32.extract_sac(data, ctable)
>>> win32.extract_sacpz(ctable)
"""

from importlib.metadata import version

from .client import Client
from .header import NETWORK

__all__ = ["Client", "NETWORK", "win32"]
# Get semantic version through setuptools-scm
__version__ = f'v{version("HinetPy")}'  # e.g. v0.1.2.dev3+g0ab3cd78



"""
Utility functions.
"""

from __future__ import annotations

import math
import shutil
from datetime import date, datetime
from importlib.metadata import version

import requests
from packaging.version import Version


def split_integer(number: int, maxn: int) -> list[int]:
    """
    Split an integer into evenly sized chunks.

    Parameters
    ----------
    number
        An interger that to be split into chunks.
    maxn
        The maximum number in each chunk.

    Returns
    -------
    chunks
        List of integers.

    Examples
    --------
    >>> split_integer(12, 3)
    [3, 3, 3, 3]
    >>> split_integer(15, 4)
    [4, 4, 4, 3]
    """
    count = math.ceil(number / maxn)
    chunks = [number // count for i in range(count)]
    for i in range(number % count):
        chunks[i] += 1
    return chunks


def point_inside_box(
    latitude: float,
    longitude: float,
    minlatitude: float | None = None,
    maxlatitude: float | None = None,
    minlongitude: float | None = None,
    maxlongitude: float | None = None,
) -> bool:
    """
    Check if a point is inside a box region.

    Parameters
    ----------
    latitude
        Latitude of the point.
    longitude
        Longitude of the point.
    minlatitude
        Minimum latitude of the box region.
    maxlatitude
        Maximum latitude of the box region.
    minlongitude
        Minimum longitude of the box region.
    maxlongitude
        Maximum longitude of the box region.

    Returns
    -------
    bool
        True if the point is inside the box region.

    Examples
    --------
    >>> point_inside_box(40, 130)
    True
    >>> point_inside_box(40, 130, 0, 50, 100, 150)
    True
    >>> point_inside_box(40, 130, 0, 30, 100, 150)
    False
    >>> point_inside_box(40, -130, maxlongitude=150)
    False
    >>> point_inside_box(40, -130, maxlongitude=300)
    True
    """
    if (minlongitude and minlongitude < 0.0) or (maxlongitude and maxlongitude < 0.0):
        raise ValueError("minlongitude and maxlongitude should be in 0-360.")
    longitude = longitude + 360.0 if longitude < 0.0 else longitude

    if minlatitude and latitude < minlatitude:
        return False
    if maxlatitude and latitude > maxlatitude:
        return False
    if minlongitude and longitude < minlongitude:
        return False
    return not (maxlongitude and longitude > maxlongitude)


def haversine(lat1: float, lon1: float, lat2: float, lon2: float) -> float:
    """
    Calculate the great circle distance between two points on the earth (specified in
    decimal degrees) using haversine formula.

    Reference: https://stackoverflow.com/a/4913653/7770208.

    >>> haversine(40, 130, 50, 140)
    12.224069629545902
    >>> haversine(-20, 50, 30, 70)
    53.57930271469817
    """
    # convert decimal degrees to radians
    lon1, lat1, lon2, lat2 = map(math.radians, [lon1, lat1, lon2, lat2])

    # haversine formula
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    delta = (
        math.sin(dlat / 2.0) ** 2.0
        + math.cos(lat1) * math.cos(lat2) * math.sin(dlon / 2.0) ** 2.0
    )
    return 2.0 * math.degrees(math.asin(math.sqrt(delta)))


def point_inside_circular(
    lat1: float,
    lon1: float,
    lat2: float,
    lon2: float,
    minradius: float | None = None,
    maxradius: float | None = None,
) -> bool:
    """
    Check if a point is inside a circular region.

    Parameters
    ----------
    lat1
        Latitude of the point.
    lon1
        Longitude of the point.
    lat2
        Latitude of center of the circular region.
    lon2
        Longitude of center of the circular region.
    minradius
        Minimum radius in degrees of the circular region.
    maxradius
        Maximum radius in degrees of the circular region.

    Returns
    -------
    bool
        True if the point is inside the circular region.

    Examples
    --------
    >>> point_inside_circular(30, 50, 30, 52, 0, 5)
    True
    """
    radius = haversine(lat1, lon1, lat2, lon2)
    return not (
        (minradius and radius < minradius) or (maxradius and radius > maxradius)
    )


def to_datetime(value: str | datetime | date) -> datetime:
    """
    Convert a datetime from :class:`str` to :class:`datetime.datetime` in a hard way.

    Parameters
    ----------
    value
        A :class:`datetime.datetime` object or a datetime string.

    Returns
    -------
    datetime
        A datetime as :class:`datetime.datetime`.

    Examples
    --------
    >>> to_datetime("201001010000")
    datetime.datetime(2010, 1, 1, 0, 0)
    >>> to_datetime("2010-01-01T03:45")
    datetime.datetime(2010, 1, 1, 3, 45)
    """
    # is datetime
    if isinstance(value, datetime):
        return value
    # is date
    if isinstance(value, date):
        return datetime.combine(value, datetime.min.time())

    # is a string
    for char in ["T", "-", ":", ",", "_"]:
        value = value.replace(char, " ")

    parts = value.split(" ")
    strfmt = "%Y%m%d%H%M%S"
    if len(parts) == 1:
        if len(value) == 8:
            strfmt = "%Y%m%d"
        elif len(value) == 12:
            strfmt = "%Y%m%d%H%M"
        elif len(value) == 14:
            strfmt = "%Y%m%d%H%M%S"
        elif len(value) > 14:
            strfmt = "%Y%m%d%H%M%S.%f"
    elif len(parts) == 3:
        strfmt = "%Y %m %d"
    elif len(parts) == 5:
        strfmt = "%Y %m %d %H %M"
    elif len(parts) == 6:
        strfmt = "%Y %m %d %H %M %S.%f" if "." in value else "%Y %m %d %H %M %S"

    return datetime.strptime(value, strfmt)


def check_cmd_exists(cmd: str) -> bool:
    """
    Check if a command exists in PATH and is executable.

    Parameters
    ----------
    cmd
        Name of the command.

    Returns
    -------
    bool
        True is the command exists in PATH is executable.
    """
    fullpath = shutil.which(cmd)
    if fullpath:
        print(f"{cmd}: Full path is {fullpath}.")
    else:
        print(f"{cmd}: Not found in PATH or isn't executable.")
    return bool(fullpath)


def check_package_release() -> bool:
    """
    Check whether HinetPy has a new release.

    Returns
    -------
    bool
        True if HinetPy has a new release.
    """
    res = requests.get("https://pypi.org/pypi/HinetPy/json", timeout=30)
    if res.status_code != 200:
        raise requests.HTTPError("Error in connecting to PyPI.")
    latest_release = res.json()["info"]["version"]

    current_version = f'v{version("HinetPy")}'
    if Version(latest_release) > Version(current_version):
        print(
            f"HinetPy v{latest_release} is released. "
            "See https://pypi.org/project/HinetPy/ for details."
        )
        return True

    print(f"You're using the latest version (v{current_version}).")
    return False



"""
Client for requesting Hi-net waveform data and catalog.
"""

import csv
import json
import logging
import os
import re
import shutil
import tempfile
import time
import zipfile
from datetime import datetime, timedelta
from html.parser import HTMLParser
from multiprocessing.pool import ThreadPool

import certifi
import requests
from requests.adapters import HTTPAdapter
from urllib3 import PoolManager
from urllib3.util import create_urllib3_context

from .header import NETWORK
from .utils import (
    check_cmd_exists,
    check_package_release,
    point_inside_box,
    point_inside_circular,
    split_integer,
    to_datetime,
)
from .win32 import merge

# Setup the logger
FORMAT = "[%(asctime)s] %(levelname)s: %(message)s"
logging.basicConfig(level=logging.INFO, format=FORMAT, datefmt="%Y-%m-%d %H:%M:%S")
logger = logging.getLogger(__name__)


# Hacking solution for "ssl.SSLError: [SSL: DH_KEY_TOO_SMALL] dh key too small" error.
# Reference: https://stackoverflow.com/a/76217135
class AddedCipherAdapter(HTTPAdapter):
    def init_poolmanager(self, connections, maxsize, block=False):
        ctx = create_urllib3_context(ciphers=":HIGH:!DH:!aNULL")
        self.poolmanager = PoolManager(
            num_pools=connections, maxsize=maxsize, block=block, ssl_context=ctx
        )


class BaseClient:
    """
    Base client to login in the Hi-net website.
    """

    # Hinet website
    _HINET = "https://www.hinet.bosai.go.jp/"
    # Authorization page
    _AUTH = "https://hinetwww11.bosai.go.jp/auth/"
    # Download win32tools
    _WIN32TOOLS = _AUTH + "manual/dlDialogue.php?r=win32tools"

    # Catalog
    _JMA = _AUTH + "JMA/dlDialogue.php"

    # Continuous wavefroms
    _CONT = _AUTH + "download/cont/"
    _CONT_STATUS = _CONT + "cont_status.php"
    _CONT_REQUEST = _CONT + "cont_request.php"
    _CONT_DOWNLOAD = _CONT + "cont_download.php"

    # Station information
    _STATION_INFO = _HINET + "st_info/detail/dlDialogue.php?f=CSV"
    _CONT_SELECT = _CONT + "select_confirm.php"
    _STATION = _CONT + "select_info.php"
    _MESONET_STATION_INFO = _CONT + "st_mesonet_json.php"
    _SNET_STATION_INFO = _CONT + "st_snet_json.php"

    # Event waveforms
    _EVENT = _AUTH + "download/event/"
    _EVENT_STATUS = _EVENT + "event_status.php"
    _EVENT_REQUEST = _EVENT + "event_request.php"
    _EVENT_DOWNLOAD = _EVENT + "event_download.php"

    # ETAG for v160422
    _ETAG = "1b61-5774e12e97f00"

    def __init__(
        self,
        user=None,
        password=None,
        timeout=60,
        retries=3,
        sleep_time_in_seconds=5,
        max_sleep_count=30,
    ):
        """
        Parameters
        ----------
        user: str
            Username of Hi-net account.
        password: str
            Password of Hi-net account.
        timeout: float
            Time to wait for the server to send data before giving up.
        retries: int
            How many times to retry if a request fails.
        sleep_time_in_seconds: float
            Sleep time between each data status check. See notes below.
        max_sleep_count: int
            Maximum number of sleeps before failing. See notes below.

        Notes
        -----
        The Hi-net server ususally spends 10-60 seconds on data preparation after
        receiving a data request. During the data preparation, users are **NOT** allowed
        to post another data request. So users have to wait until the data is ready.

        HinetPy checks data status every ``sleep_time_in_seconds`` seconds for no more
        than ``max_sleep_count`` times, until the data is ready. If the data status is
        still NOT ready after ``max_sleep_count * sleep_time_in_seconds`` seconds, it
        most likely means something goes wrong with the data request. Then, HinetPy will
        retry to request the data ``retries`` times. Ususally, you don't need to modify
        these parameters unless you know what you're doing.

        Examples
        --------

        >>> from HinetPy import Client
        >>> client = Client("username", "password")
        """
        self.timeout = timeout
        self.retries = retries
        self.sleep_time_in_seconds = sleep_time_in_seconds
        self.max_sleep_count = max_sleep_count
        if user and password:
            self.login(user, password)
        # variables for internal use
        self._code = None
        self._max_span = 0

    def __str__(self):
        """String representation for the client."""
        string = "<== Hi-net web service client ==>\n"
        string += f"{'url':22s}: {self._HINET}\n"
        for key in (
            "user",
            "password",
            "timeout",
            "retries",
            "debug",
            "max_sleep_count",
            "sleep_time_in_seconds",
        ):
            try:
                value = getattr(self, key)
                if key == "password":
                    value = "*" * len(value)
                string += f"{key:22s}: {value}\n"
            except AttributeError:  # noqa: PERF203
                continue
        return string

    def login(self, user, password):
        """
        Login in the Hi-net server.

        Parameters
        ----------
        user: str
            Username of Hi-net account.
        password: str
            Password of Hi-net account.

        Examples
        --------
        >>> from HinetPy import Client
        >>> client = Client()
        >>> client.login("username", "password")
        """
        self.user = user
        self.password = password[:12]  # Hinet truncates password > 12 characters
        if len(password) > 12:
            logger.warning("Password longer than 12 characters may be truncated.")
        self.session = requests.Session()
        self.session.mount(self._HINET, AddedCipherAdapter())
        self.session.mount(self._AUTH, AddedCipherAdapter())
        self.session.get(
            self._AUTH, timeout=self.timeout, verify=certifi.where()
        )  # get cookie
        resp = self.session.post(
            self._AUTH,
            data={"auth_un": self.user, "auth_pw": self.password},
            timeout=self.timeout,
        )
        # Hi-net server returns 200 even when the username or password is wrong, thus
        # I have to check the webpage content to make sure login is successful
        if re.search(r"auth_log(?P<LOG>.*)\.png", resp.text).group("LOG") == "out":
            msg = "Unauthorized. Check your username and password!"
            raise requests.ConnectionError(msg)


class ContinuousWaveformClient(BaseClient):
    """
    Client for requesting continuous waveform data.
    """

    def _request_cont_waveform(self, code, starttime, span):
        """
        Request continuous waveform.

        Parameters
        ----------
        code: str
            Network code.
        starttime: :py:class:`datetime.datetime`
            Starttime of data request.
        span: int
            Time span in minutes.

        Returns
        -------
        id: str
            ID of the requested data. None if request fails.
        """
        org, net, volc = _parse_code(code)
        payload = {
            "org1": org,
            "org2": net,
            "volc": volc,
            "year": starttime.strftime("%Y"),
            "month": starttime.strftime("%m"),
            "day": starttime.strftime("%d"),
            "hour": starttime.strftime("%H"),
            "min": starttime.strftime("%M"),
            "span": str(span),
            "arc": "ZIP",
            "size": "93680",  # the actual size doesn't matter
            "LANG": "en",
        }

        # retry if request fails else return id
        for _ in range(self.retries):
            try:
                # the timestamp determines the unique id
                payload["rn"] = str(int(datetime.now().timestamp()))
                resp = self.session.post(
                    self._CONT_REQUEST, params=payload, timeout=self.timeout
                )
                # assume the first one on the status page is the current data
                id = re.search(  # noqa: A001
                    r'<td class="bgcolist2">(?P<ID>\d{10})</td>', resp.text
                ).group("ID")
                p = re.compile(
                    r'<tr class="bglist(?P<OPT>\d)">'  # noqa: ISC003
                    + r'<td class="bgcolist2">'
                    + id
                    + r"</td>"
                )
                # wait until data is ready
                for _i in range(self.max_sleep_count):
                    status = self.session.post(self._CONT_STATUS, timeout=self.timeout)
                    opt = p.search(status.text).group("OPT")
                    if opt == "1":  # still preparing data
                        time.sleep(self.sleep_time_in_seconds)
                    elif opt == "2":  # data is available
                        return id
                    elif opt == "3":  # ?
                        logger.error(
                            "If you see this, please report an issue on GitHub!"
                        )
                        break  # break to else clause
                    elif opt == "4":  # something wrong, retry
                        logger.error("Error in data status.")
                        break  # break to else clause
                else:  # wait too long time
                    return None
            except Exception:  # noqa: BLE001, S112
                continue
            break
        else:
            logger.error("Data request fails after %d retries", self.retries)
            return None

    def _download_cont_waveform(self, job):
        """
        Download continuous waveform.

        Parameters
        ----------
        id: str
            Data id to be downloaded.

        Returns
        -------
        cnts: list of str
            Filenames of one-minute win32 data.
        ctable: str
            Filenames of channle tables.
        """
        # session cannot be shared between threads, so always initialize
        # a new client for downloading.
        # Maybe there is other tricky way?
        dlclient = Client(self.user, self.password)
        for _ in range(self.retries):
            try:
                resp = dlclient.session.post(
                    self._CONT_DOWNLOAD,
                    data={"id": job.id},
                    stream=True,
                    timeout=self.timeout,
                )

                with tempfile.NamedTemporaryFile() as ft:
                    # save to temporary file
                    for chunk in resp.iter_content(chunk_size=1024):
                        if chunk:  # filter out keep-alive new chunks
                            ft.write(chunk)
                    ft.flush()

                    # unzip temporary file
                    cnts, ctable = [], None
                    with zipfile.ZipFile(ft.name) as fz:
                        for filename in fz.namelist():
                            if filename.endswith(".cnt"):
                                cnts.append(filename)
                            elif filename.endswith(".euc.ch"):
                                ctable = filename
                        fz.extractall(members=[*cnts, ctable])
                    return cnts, ctable
            except Exception:  # noqa: BLE001, S112
                continue
            break
        else:
            logger.error("Data download fails after %d retries", self.retries)
            return None, None

    def get_continuous_waveform(  # noqa: PLR0915, PLR0912
        self,
        code,
        starttime,
        span,
        max_span=None,
        data=None,
        ctable=None,
        outdir=None,
        threads=3,
        cleanup=True,
    ):
        """
        Get continuous waveform data from Hi-net server.

        Parameters
        ----------
        code: str
            Network code. See :meth:`~HinetPy.client.Client.info` for details.
        starttime: :py:class:`datetime.datetime` or str
            Starttime of a data request.
        span: int
            Time span in minutes.
        max_span: int
            Maximum time span for sub-requests. Defaults to be determined automatically.
            See notes below.
        data: str
            Filename of downloaded win32 data.
            Default format: CODE_YYYYmmddHHMM_SPAN.cnt
        ctable: str
            Filename of downloaded channel table file. Default format: CODE_YYYYmmdd.ch
        outdir: str
            Save win32 and channel table data to a specified directory. Default is in
            the current directory.
        threads: int
            Parallel data download using more threads.
        cleanup: bool
            Clean up one-minute cnt files after merging.

        Returns
        -------
        data: str
            Filename of downloaded win32 data.
        ctable: str
            Filename of downloaded channel table file.

        Notes
        -----
        **TimeZone**

        All times in HinetPy and in Hi-net data are in JST (UTC+0900).

        **max_span**

        Hi-net server sets three limitations for each data request:

        1. Record_Length <= 60 min
        2. Number_of_channels * Record_Length <= 12000 min
        3. Only the latest 150 requested data are kept

        For example, Hi-net network has about 24000 channels. Acoording to limitation 2,
        the record length should be no more than 5 minutes for each data request.
        HinetPy "break through" the limitation by splitting a long-duration data request
        into several short-duration sub-requsts.

        **How it works**

        1. do several checks
        2. split a long-duration request into several short-duration sub-requests
        3. loop over all sub-requests and return the data IDs
        4. download all data based on the data IDs
        5. extract all zip files and merge them into one single win32 data
        6. clean up

        Examples
        --------
        Request 6-minute data since 2010-01-01T05:35 (UTC+0900) from Hi-net.

        >>> client.get_continuous_waveform("0101", "201001010535", 6)
        ('0101_201001010535_6.cnt', '0101_20100101.ch')

        Several other string formats of ``starttime`` are also supported:

        >>> client.get_continuous_waveform("0101", "2010-01-01 05:35", 6)
        >>> client.get_continuous_waveform("0101", "2010-01-01T05:35", 6)

        :py:class:`datetime.datetime` is also supported:

        >>> from datetime import datetime
        >>> starttime = datetime(2010, 1, 1, 5, 35)
        >>> client.get_continuous_waveform("0101", starttime, 6)
        ('0101_201001010535_6.cnt', '0101_20100101.ch')

        Request full-day data of 2010-01-01T00:00 (UTC+0900) of F-net:

        >>> client.get_continuous_waveform("0103", starttime, 1440, max_span=25)
        ('0103_201001010000_1440.cnt', '0103_20100101.ch')

        """
        # 1. check span:
        #    max limit is determined by the max number of data points
        #    allowed in code s4win2sacm.c
        if not isinstance(span, int):
            raise TypeError("span must be integer.")
        if not 1 <= span <= (2**31 - 1) / 6000:
            raise ValueError("Span is NOT in the allowed range [1, 357913]")

        # 2. check starttime and endtime
        if code not in NETWORK:
            raise ValueError(f"{code}: Incorrect network code.")

        starttime = to_datetime(starttime)
        endtime = starttime + timedelta(minutes=span)

        # Special handling for MeSO-net since it has two network codes:
        # - 0131: Data after 20170401
        # - 0231: Data 20080516-20170401
        mesonet_time = NETWORK["0131"].starttime
        if code == "0131" and starttime < mesonet_time:
            code = "0231"
        if code == "0231" and starttime >= mesonet_time:
            code = "0131"
        if code in ("0131", "0231") and starttime <= mesonet_time <= endtime:
            msg = (
                "MeSO-net has two network codes: '0131' for data after 20170401, and "
                "'0231' for data 20080516-20170401. But the time span you requested "
                "crosses the date 20170401. Please use the correct network code."
            )
            raise ValueError(msg)

        time0 = NETWORK[code].starttime
        # time1 = UTCTime + JST (UTC+0900) - 2 hour delay
        time1 = datetime.utcnow() + timedelta(hours=9) + timedelta(hours=-2)
        if not time0 <= starttime < endtime <= time1:
            msg = f"Data not available in the time period. Call Client.info('{code}') for help."
            raise ValueError(msg)

        # 3. set max_span
        if self._code != code:  # update default max_span
            self._code = code
            self._max_span = self._get_allowed_span(code)
        if not (max_span and 1 <= max_span <= 60):
            max_span = self._max_span

        # 4. prepare jobs
        jobs = prepare_jobs(starttime, span, max_span)

        cnts, ch_euc = [], set()
        logger.info("%s ~%s", starttime.strftime("%Y-%m-%d %H:%M"), span)
        # 5. request and download
        count = len(jobs)
        for j in range(0, count, 100):  # to break the limitation of 150
            # 5.1. request <=100 data
            for i in range(j, min(j + 100, count)):
                logger.info(
                    "[%s/%d] => %s ~%d",
                    str(i + 1).zfill(len(str(count))),
                    count,
                    jobs[i].starttime.strftime("%Y-%m-%d %H:%M"),
                    jobs[i].span,
                )
                jobs[i].id = self._request_cont_waveform(
                    code, jobs[i].starttime, jobs[i].span
                )

            # 5.2. check ids
            if not [job.id for job in jobs]:
                logger.error("No data requested succesuflly. Skipped.")
                return None, None
            # check if all ids are not None
            if not all(job.id for job in jobs):
                logger.error("Fail to request some data. Skipped.")
                return None, None

            # 5.3. parallel downloading
            with ThreadPool(min(threads, len(jobs))) as p:
                rvalue = p.map(self._download_cont_waveform, jobs)
            for value in rvalue:
                cnts.extend(value[0])
                ch_euc.add(value[1])

        # post processes
        # 1. always sort cnts by name/time to avoid use -s option of catwin32
        cnts = sorted(cnts)
        #    always use the first ctable
        ch_euc = sorted(ch_euc)[0]

        # 2. merge all cnt files
        if not data:
            data = f'{code}_{starttime.strftime("%Y%m%d%H%M")}_{span:d}.cnt'
        dirname = None
        if os.path.dirname(data):
            dirname = os.path.dirname(data)
        elif outdir:
            dirname = outdir
            data = os.path.join(dirname, data)
        merge(cnts, data)

        # 3. rename channeltable file
        if not ctable:
            ctable = f'{code}_{starttime.strftime("%Y%m%d")}.ch'

        dirname = None
        if os.path.dirname(ctable):
            dirname = os.path.dirname(ctable)
        elif outdir:
            dirname = outdir
            ctable = os.path.join(dirname, ctable)
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname, exist_ok=True)
        shutil.move(ch_euc, ctable)
        # 4. cleanup
        if cleanup:
            for cnt in cnts:
                os.remove(cnt)

        return data, ctable

    def get_waveform(self, code, starttime, span, **kwargs):
        """
        .. deprecated:: 0.6.0

            :meth:`~HinetPy.client.Client.get_waveform` is deprecated.
            Use :meth:`~HinetPy.client.Client.get_continuous_waveform` instead.
        """
        logger.warning(
            "The get_waveform() function is deprecated. "
            "Use get_continuous_waveform() instead."
        )
        return self.get_continuous_waveform(code, starttime, span, **kwargs)


class EventWaveformClient(BaseClient):
    """Client for requesting event waveform data."""

    def _search_event_by_day(
        self,
        year,
        month,
        day,
        region="00",
        magmin=3.0,
        magmax=9.9,
        include_unknown_mag=True,
    ):
        """
        Search event catalog of one day.

        Parameters
        ----------
        region : str
            Limit events in specified region. Allowed values are:

            - ``00``: Whole Japan
            - ``01``: Hokkaido Region
            - ``02``: Tohoku Region
            - ``03``: Kanto Region
            - ``04``: Chubu Region
            - ``05``: Kinki Region
            - ``06``: Chugoku/Shikoku Region
            - ``07``: Kyushu Region
            - ``08``: Others

        include_unknown_mag: bool
            Include/exclude undetermined magnitude events.
        """
        payload = {
            "year": year,
            "month": f"{month:02d}",
            "day": f"{day:02d}",
            "region": region,
            "mags": str(f"{magmin:.1f}"),  # magnitude must be 3.0, not 3
            "mage": str(f"{magmax:.1f}"),
            "undet": 0 if include_unknown_mag else 1,
            "sort": 0,  # always sort by origin time in ascending order
            "arc": "ZIP",  # meaningless arguement in this request
            "go": 1,
            "LANG": "en",
        }
        resp = self.session.post(self._EVENT, data=payload, timeout=self.timeout)
        events = []
        for result in re.findall(r"openRequest\((.+)\)", resp.text):
            items = [item.strip("'") for item in result.split(",")]
            events.append(Event(items[0], *items[3:10]))
        return events

    def _request_event_waveform(self, event, format="ZIP"):  # noqa: A002
        """
        Request event waveform.

        Parameters
        ----------
        event: :class:`HinetPy.client.Event`
            Event to be requested.
        format: str
            Format of requested waveform package.
        """
        payload = {
            "evid": event.evid,
            "origin_jst": event.origin,
            "hypo_latitude": event.latitude,
            "hypo_logitude": event.longitude,  # logitude is Hinet's typo
            "hypo_depth": event.depth,
            "mg": "" if event.magnitude == "#" else event.magnitude,
            "hypo_name": event.name,
            "hypo_name_eng": event.name_en,
            "arc": format,
            "encoding": "D",
            "lang": "en",
            "rn": str(int(datetime.now().timestamp())),
        }

        # retry if request fails else return id
        for _ in range(self.retries):
            try:
                resp = self.session.post(
                    self._EVENT_REQUEST, data=payload, timeout=self.timeout
                )
                # assume the first one on the status page is the current one
                id = re.search(  # noqa: A001
                    r'<td class="bgevlist2">(?P<ID>\d{10})</td>', resp.text
                ).group("ID")
                p = re.compile(
                    r'<tr class="bglist(?P<OPT>\d)">'  # noqa: ISC003
                    + r'<td class="bgevlist2">'
                    + id
                    + r"</td>"
                )
                # wait until data is ready
                for _i in range(self.max_sleep_count):
                    status = self.session.post(self._EVENT_STATUS, timeout=self.timeout)
                    opt = p.search(status.text).group("OPT")
                    if opt == "1":  # still preparing data
                        time.sleep(self.sleep_time_in_seconds)
                    elif opt == "2":  # data is available
                        return id
                    elif opt == "3":  # ?
                        logger.error(
                            "If you see this, please report an issue on GitHub!"
                        )
                        break  # break to else clause
                    elif opt == "4":  # something wrong, retry
                        logger.error("Error in data status.")
                        break  # break to else clause
                else:  # wait too long time
                    return None
            except Exception:  # noqa: BLE001, S112
                continue
            break
        else:
            logger.error("Data request fails after %d retries.", self.retries)
            return None

    def _download_event_waveform(self, id):  # noqa: A002
        """
        Download event waveform.

        Parameters
        ----------
        id: str
            Request ID.
        """

        dlclient = Client(self.user, self.password)
        for _ in range(self.retries):
            try:
                resp = dlclient.session.get(
                    self._EVENT_DOWNLOAD,
                    params={"id": id, "encode": "D", "LANG": "en"},
                    stream=True,
                    timeout=self.timeout,
                )
                fname = resp.headers["Content-Disposition"].split("=")[1].strip('"')
                outdir = "_".join(fname.split("_")[:2])

                with tempfile.NamedTemporaryFile() as ft:
                    # save to temporary file
                    for chunk in resp.iter_content(chunk_size=1024):
                        if chunk:  # filter out keep-alive new chunks
                            ft.write(chunk)
                    ft.flush()

                    # unzip temporary file
                    with zipfile.ZipFile(ft.name) as fz:
                        fz.extractall(path=outdir)
                    return outdir
            except Exception:  # noqa: BLE001, S112, PERF203
                continue
        logger.error("Data download fails after %d retries.", self.retries)
        return None

    def get_event_waveform(  # noqa: PLR0913
        self,
        starttime,
        endtime,
        region="00",
        minmagnitude=3.0,
        maxmagnitude=9.9,
        include_unknown_mag=True,
        mindepth=None,
        maxdepth=None,
        minlatitude=None,
        maxlatitude=None,
        minlongitude=None,
        maxlongitude=None,
        latitude=None,
        longitude=None,
        minradius=None,
        maxradius=None,
    ):
        """
        Get event waveform data.

        Parameters
        ----------
        starttime: :py:class:`datetime.datetime` or str
            Starttime of events.
        endtime: :py:class:`datetime.datetime` or str
            Endtime of events.
        region: str
            Limit events in specified region. Allowed values are:

            - ``00``: Whole Japan
            - ``01``: Hokkaido Region
            - ``02``: Tohoku Region
            - ``03``: Kanto Region
            - ``04``: Chubu Region
            - ``05``: Kinki Region
            - ``06``: Chugoku/Shikoku Region
            - ``07``: Kyushu Region
            - ``08``: Others
        minmagnitude: float
            Limit to events with a magnitude larger than speicified minimum.
        maxmagnitude: float
            Limit to events with a magnitude smaller than speicified maximum.
        include_unknown_mag: bool
            Include/exclude undetermined magnitude events.
        mindepth: float
            Limit to events deeper than the specified minimum.
        maxdepth: float
            Limit to events shallower than the specified maximum.
        minlatitude: float
            Limit to events with a latitude larger than the specified minimum.
        maxlatitude: float
            Limit to events with a latitude smaller than the specified maximum.
        minlongitude: float
            Limit to events with a longitude larger than the specified minimum.
        maxlongitude: float
            Limit to events with a longitude smaller than the specified maximum.
        latitude: float
            Specify the latitude to be used for a radius search.
        longitude: float
            Specify the longitude to be used for a radius search.
        minradius: float
            Limit to events within the specified minimum number of degrees from the
            geographic point defined by the latitude and longitude parameters.
        maxradius: float
            Limit to events within the specified maximum number of degrees from the
            geographic point defined by the latitude and longitude parameters.
        """
        starttime, endtime = to_datetime(starttime), to_datetime(endtime)

        # get event list
        events = []
        for i in range((endtime.date() - starttime.date()).days + 1):
            event_date = starttime.date() + timedelta(days=i)
            events.extend(
                self._search_event_by_day(
                    event_date.year,
                    event_date.month,
                    event_date.day,
                    region=region,
                    magmin=minmagnitude,
                    magmax=maxmagnitude,
                    include_unknown_mag=include_unknown_mag,
                )
            )

        # select events
        selected_events = []
        for event in events:
            # select events based on origin time
            if not starttime <= event.origin <= endtime:
                continue
            # select events based on magnitude
            if (event.magnitude != -99.9) and (
                not minmagnitude <= event.magnitude <= maxmagnitude
            ):
                continue
            # select events based on depth
            if mindepth is not None and event.depth < mindepth:
                continue
            if maxdepth is not None and event.depth > maxdepth:
                continue

            # select events in a box region
            if (
                minlatitude is not None
                or maxlatitude is not None
                or minlongitude is not None
                or maxlongitude is not None
            ) and not point_inside_box(
                event.latitude,
                event.longitude,
                minlatitude=minlatitude,
                maxlatitude=maxlatitude,
                minlongitude=minlongitude,
                maxlongitude=maxlongitude,
            ):
                continue

            # select events in a circular region
            if (
                (latitude is not None and longitude is not None)
                and (minradius is not None or maxradius is not None)
                and not point_inside_circular(
                    event.latitude,
                    event.longitude,
                    latitude,
                    longitude,
                    minradius=minradius,
                    maxradius=maxradius,
                )
            ):
                continue
            selected_events.append(event)

        logger.info("EVENT WAVEFORM DOWNLOADER:")
        logger.info("%d events to download.", len(selected_events))
        for event in selected_events:
            id = self._request_event_waveform(event)  # noqa: A001
            dirname = self._download_event_waveform(id)
            logger.info("%s %s", event, dirname)


class CatalogClient(BaseClient):
    """Client for requesting catalogs."""

    def _get_catalog(self, datatype, startdate, span, filename=None, os="DOS"):
        """Request JMA catalog."""

        startdate = to_datetime(startdate)
        if int(span) not in range(1, 8):
            raise ValueError("span is not digit or not in [1, 7].")

        params = {
            "data": datatype,
            "rtm": startdate.strftime("%Y%m%d"),
            "span": span,
            "os": os[0],
        }
        d = self.session.post(self._JMA, params=params, stream=True)
        if not filename:
            filename = f'{datatype}_{startdate.strftime("%Y%m%d")}_{span}.txt'
        with open(filename, "wb") as fd:
            for chunk in d.iter_content(chunk_size=1024):
                if chunk:  # filter out keep-alive new chunks
                    fd.write(chunk)
        return filename

    def get_arrivaltime(self, startdate, span, filename=None, os="DOS"):
        """
        Get JMA arrival time data from Hi-net.

        Parameters
        ----------
        startdate: str, :py:class:`datetime.date`, :py:class:`datetime.datetime`
            Start date to request.
        span: int
            Data length in days.
        os: str
            File format. "DOS" or "UNIX".
        filename: str
            Filename to save.

        Returns
        -------
        str
            Filename saved.

        Examples
        --------
        >>> from datetime import date
        >>> startdate = date(2010, 1, 1)
        >>> client.get_arrivaltime(startdate, 5)
        'measure_20100101_5.txt'
        >>> client.get_arrivaltime(startdate, 5, filename="arrivaltime.txt")
        'arrivaltime.txt'
        """
        return self._get_catalog("measure", startdate, span, filename, os)

    def get_focalmechanism(self, startdate, span, filename=None, os="DOS"):
        """
        Get JMA focal mechanism data from Hi-net.

        Parameters
        ----------
        startdate: str, :py:class:`datetime.date`, :py:class:`datetime.datetime`
            Start date to request.
        span: int
            Data length in days.
        os: str
            File format. "DOS" or "UNIX".
        filename: str
            Filename to save.

        Returns
        -------
        str
            Filename saved.

        Examples
        --------
        >>> from datetime import date
        >>> startdate = date(2010, 1, 1)
        >>> client.get_focalmechanism(startdate, 5)
        'focal_20100101_5.txt'
        >>> client.get_focalmechanism(startdate, 5, filename="focal.txt")
        'focal.txt'
        """
        return self._get_catalog("mecha", startdate, span, filename, os)


class StationClient(BaseClient):
    """
    Client for manipulating stations.
    """

    def get_station_list(self, code):
        """
        Get station list of a network.

        The function only supports the following networks:
        Hi-net (0101), F-net (0103, 0103A), S-net (0120, 0120A) and MeSO-net (0131).

        >>> stations = client.get_station_list("0101")
        >>> for station in stations:
        ...     print(station)
        0101 N.WNNH 45.4883 141.885 -159.06
        0101 N.SFNH 45.3346 142.1185 -81.6
        ...
        """
        stations = []
        # remove trailing 'A' in network code
        if code in ["0101", "0103", "0103A"]:  # Hinet and Fnet
            csvfile = requests.get(self._STATION_INFO, timeout=30).content.decode(
                "utf-8"
            )
            for row in csv.DictReader(csvfile.splitlines(), delimiter=","):
                org_id = row["organization_id"].strip("'")
                net_id = row["network_id"].strip("'")
                if org_id + net_id != code[:4]:
                    continue
                name = row["station_cd"]
                latitude, longitude = row["latitude"], row["longitude"]
                elevation = row["height(m)"]
                stations.append(Station(code, name, latitude, longitude, elevation))
        elif code in ["0120", "0120A", "0131"]:  # S-net and MeSO-net
            if code in ["0120", "0120A"]:
                url = self._SNET_STATION_INFO
                ltext, rtext = "var snet_station = [", "];"
            else:
                url = self._MESONET_STATION_INFO
                ltext, rtext = "var mesonet_station = [", "];"
            json_text = self.session.get(url).text.lstrip(ltext).rstrip(rtext)
            for station in json.loads(json_text)["features"]:
                prop = station["properties"]
                name = prop["station_cd"]
                latitude, longitude = prop["latitude"], prop["longitude"]
                elevation = prop["sensor_height"]
                stations.append(Station(code, name, latitude, longitude, elevation))
        else:
            raise ValueError("Only support Hi-net, F-net, S-net and MeSO-net.")
        return stations

    def get_selected_stations(self, code: str):
        """
        Query stations selected for requesting data.

        It supports two networks: Hi-net (0101) and F-net (0103, 0103A).

        Parameters
        ----------
        code
            Network code.

        Returns
        -------
        stations: list of :class:`~HinetPy.client.Station`
            List of selected stations with station metadata information.

        Examples
        --------
        >>> stations = client.get_selected_stations("0101")
        >>> len(stations)
        16
        >>> for station in stations:
        ...     print(station)
        0101 N.WNNH 45.4883 141.885 -159.06
        0101 N.SFNH 45.3346 142.1185 -81.6
        >>> names = [station.name for station in stations]
        >>> print(*names)
        N.WNNH N.SFNH ...
        """
        if code == "0101":
            pattern = r"N\..{3}H"
        elif code in ("0103", "0103A"):
            pattern = r"N\..{3}F"
        else:
            raise ValueError("Can only query stations of Hi-net/F-net")

        parser = _GrepTableData()
        parser.feed(self.session.get(self._STATION, timeout=self.timeout).text)
        stations = []
        for i, text in enumerate(parser.tabledata):
            # If the target station, grep both lon and lat.
            if re.match(pattern, text):
                latitude, longitude, elevation = parser.tabledata[i + 3 : i + 6]
                latitude = latitude.strip("N")
                longitude = longitude.strip("E")
                elevation = elevation.strip("m")

                stations.append(
                    Station(
                        code=code,
                        name=text,
                        latitude=latitude,
                        longitude=longitude,
                        elevation=elevation,
                    )
                )
        parser.close()
        return stations

    def select_stations(
        self,
        code,
        stations=None,
        minlatitude=None,
        maxlatitude=None,
        minlongitude=None,
        maxlongitude=None,
        latitude=None,
        longitude=None,
        minradius=None,
        maxradius=None,
    ):
        """
        Select stations of a network.

        It only supports the following networks:
        Hi-net (0101), F-net (0103, 0103A), S-net (0120, 0120A) and MeSO-net (0131).

        Parameters
        ----------
        code: str
            Network code.
        stations: str or list
            Stations to select.
        minlatitude: float
            Limit to stations with a latitude larger than the specified minimum.
        maxlatitude: float
            Limit to stations with a latitude smaller than the specified maximum.
        minlongitude: float
            Limit to stations with a longitude larger than the specified minimum.
        maxlongitude: float
            Limit to stations with a longitude smaller than the specified maximum.
        latitude: float
            Specify the latitude to be used for a radius search.
        longitude: float
            Specify the longitude to be used for a radius search.
        minradius: float
            Limit to stations within the specified minimum number of degrees from the
            geographic point defined by the latitude and longitude parameters.
        maxradius: float
            Limit to stations within the specified maximum number of degrees from the
            geographic point defined by the latitude and longitude parameters.

        Examples
        --------
        Select only two stations of Hi-net:

        >>> client.select_stations("0101", ["N.AAKH", "N.ABNH"])
        >>> len(client.get_selected_stations("0101"))
        2

        Select stations in a box region:

        >>> client.select_stations(
        ...     "0101",
        ...     minlatitude=40,
        ...     maxlatitude=50,
        ...     minlongitude=140,
        ...     maxlongitude=150,
        ... )

        Select stations in a circular region:

        >>> client.select_stations(
        ...     "0101", latitude=30, longitude=139, minradius=0, maxradius=2
        ... )

        Select all Hi-net stations:

        >>> client.select_stations("0101")
        >>> len(client.get_selected_stations("0101"))
        0

        """
        stations_selected = []

        if stations is None:
            pass
        elif isinstance(stations, str):  # stations is a str, i.e., one station
            stations_selected.append(stations)
        elif isinstance(stations, list):  # list of stations
            stations_selected.extend(stations)
        else:
            raise ValueError("stations should be either a str or a list.")

        # get station list from Hi-net server
        stations_at_server = self.get_station_list(code)

        # select stations in a box region
        if minlatitude or maxlatitude or minlongitude or maxlongitude:
            stations_selected = [
                station.name
                for station in stations_at_server
                if station.code == code
                and point_inside_box(
                    station.latitude,
                    station.longitude,
                    minlatitude=minlatitude,
                    maxlatitude=maxlatitude,
                    minlongitude=minlongitude,
                    maxlongitude=maxlongitude,
                )
            ]

        # select stations in a circular region
        if (latitude and longitude) and (minradius or maxradius):
            stations_selected = [
                station.name
                for station in stations_at_server
                if station.code == code
                and point_inside_circular(
                    station.latitude,
                    station.longitude,
                    latitude,
                    longitude,
                    minradius=minradius,
                    maxradius=maxradius,
                )
            ]
        payload = {
            "net": code,
            "stcds": ":".join(stations_selected) if stations_selected else None,
            "mode": "1",
        }
        self.session.post(self._CONT_SELECT, data=payload, timeout=self.timeout)


class Client(
    ContinuousWaveformClient, EventWaveformClient, CatalogClient, StationClient
):
    """
    Wrapper client to request waveform, catalog and manipulate stations.
    """

    def doctor(self):
        """
        Doctor does some checks.

        This is a utility function that checks:

        - if HinetPy has a new release
        - if Hi-net web service is updated
        - if ``catwin32`` and ``win2sac_32`` are in PATH

        >>> client.doctor()
        Hi-net web service is NOT updated.
        You're using the latest release (v0.x.x).
        catwin32: Full path is /home/user/bin/catwin32.
        win2sac_32: Full path is /home/user/bin/win2sac_32.
        """
        self.check_service_update()
        check_package_release()
        check_cmd_exists("catwin32")
        check_cmd_exists("win2sac_32")

    def _get_allowed_span(self, code):
        """
        Get allowed max span for each network.

        Hi-net server sets two limitations of data file size:

        #. Number_of_channels * record_length(min.) <= 12000 min
        #. record_length <= 60min

        >>> client._get_allowed_span("0201")
        60

        Parameters
        ----------
        code: str
            Network code.

        Returns
        -------
        max_span: int
            Maximum allowed span in mimutes.
        """
        # hard-coded total number of channels
        channels = NETWORK[code].channels
        # query the actual number of channels
        if code in ("0101", "0103", "0103A"):
            stations = len(self.get_selected_stations(code))
            if stations != 0:
                channels = stations * 3

        if code in ("0103", "0103A"):
            # Maximum allowed file size is ~55 MB for F-net
            f_net_dl_factor, f_net_max_size = 8.8667638012, 55000
            return int(f_net_max_size / (f_net_dl_factor * channels))
        return min(int(12000 / channels), 60)

    def check_service_update(self):
        """
        Check if Hi-net service is updated.

        >>> client.check_service_update()
        [2017-01-01 00:00:00] INFO: Hi-net web service is NOT updated.
        """
        resp = self.session.get(self._CONT + "/js/cont.js")
        if resp.headers["ETag"].strip('"') == self._ETAG:
            logger.info("Hi-net web service is NOT updated.")
            return False
        logger.warning("Hi-net web service is updated. HinetPy may FAIL!")
        return True

    def info(self, code=None):
        """
        Show information of networks.

        Parameters
        ----------
        code: None or str
            Network code.

        Examples
        --------

        >>> client.info()
        0101   : NIED Hi-net
        0103   : NIED F-net (broadband)
        0103A  : NIED F-net (strong motion)
        010501 : NIED V-net (Tokachidake)
        ...
        0703   : Aomori Prefectural Government
        0705   : Shizuoka Prefectural Government
        0801   : ADEP
        >>> client.info("0101")
        == Information of Network 0101 ==
        Name: NIED Hi-net
        Starttime: 20040401
        No. of channels: 2336
        """
        if code:
            net = NETWORK[code]
            info = f"== Information of Network {code} ==\n"
            info += f"Name: {net.name}\n"
            info += f"Homepage: {net.url}\n"
            info += f"Starttime: {net.starttime.strftime('%Y%m%d')}\n"
            info += f"No. of channels: {net.channels}"
            print(info)
        else:
            for network_code in sorted(NETWORK):
                print(f"{network_code:7s}: {NETWORK[network_code].name}")

    def _get_win32tools(self):
        """Download win32tools from Hi-net website."""
        dl = self.session.get(self._WIN32TOOLS, stream=True)
        if dl.status_code != 200:
            logger.error("Error in downloading win32tools.")
        with open("win32tools.tar.gz", "wb") as fd:
            for chunk in dl.iter_content(chunk_size=1024):
                if chunk:  # filter out keep-alive new chunks
                    fd.write(chunk)


def prepare_jobs(starttime, span, max_span):
    """Prepare jobs."""
    spans = split_integer(span, max_span)
    jobs = [_Job(starttime=starttime, span=spans[0])]
    for i in range(1, len(spans)):
        dt = jobs[i - 1].starttime + timedelta(minutes=spans[i - 1])
        jobs.append(_Job(dt, spans[i]))
    return jobs


class _Job:
    """Job class for internal use."""

    def __init__(self, starttime, span, id=None):  # noqa: A002
        self.starttime = starttime
        self.span = span
        self.id = id


class Station:
    """Class for Stations."""

    def __init__(self, code, name, latitude, longitude, elevation):
        self.code = code
        self.name = name
        self.latitude = float(latitude) if latitude else None
        self.longitude = float(longitude) if longitude else None
        self.elevation = float(elevation) if elevation else None

    def __str__(self):
        return (
            f"{self.code} {self.name} {self.latitude} {self.longitude} {self.elevation}"
        )


class Event:
    """Event class for requesting event waveforms."""

    def __init__(
        self, evid, origin, latitude, longitude, depth, magnitude, name, name_en
    ):
        self.evid = evid
        self.origin = datetime.strptime(origin, "%Y/%m/%d %H:%M:%S.%f")
        if latitude[-1] == "N":
            self.latitude = float(latitude[:-1])
        else:
            self.latitude = -float(latitude[:-1])

        if longitude[-1] == "E":
            self.longitude = float(longitude[:-1])
        else:
            self.longitude = -float(longitude[:-1])

        self.depth = float(depth.strip("km"))
        if magnitude == "#":  # unknown magnitude
            self.magnitude = -99.9
        else:
            self.magnitude = float(magnitude)
        self.name = name
        self.name_en = name_en

    def __str__(self):
        return (
            f"{self.origin} {self.latitude} {self.longitude}"
            f" {self.depth} {self.magnitude}"
        )


def _parse_code(code):
    """
    Parse network code.

    >>> client._parse_code("0101")
    ('01', '01', '0')
    >>> client._parse_code("0103A")
    ('01', '03A', '0')
    >>> client._parse_code("010501")
    ('01', '05', '010501')
    """
    if code not in NETWORK:
        raise ValueError(f"{code}: Incorrect network code.")

    if code.startswith(("0105", "0302")):
        org, net, volc = code[0:2], code[2:4], code
    else:
        org, net, volc = code[:2], code[2:], "0"
    return org, net, volc


class _GrepTableData(HTMLParser):
    """
    Parser to obtain ``<td>`` contents.
    ``handle_starttag()`` flags when the HTML tag matches with `td`.
    """

    def __init__(self):
        super().__init__()
        self.if_tabledata = False
        self.tabledata = []

    def handle_starttag(self, tag, attrs):
        if re.match("^td$", tag):
            self.if_tabledata = True

    def handle_data(self, data):
        if self.if_tabledata:
            self.tabledata.append(data)
            self.if_tabledata = False
